﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=6.3
@EndOfDesignText@
'version 0.94
Sub Class_Globals
	Private xui As XUI
	Private cache As Map
	Type X2SpriteGraphicData (Name As String, MapsOfCompressedBCs As List, AntiAlias As Boolean, AngleInterval As Int, _
		OriginalBMPs As List, OriginalBCs As List, SizeOfAllCompressed As Int, LastUsed As Long, VerticalSymmetry As Boolean, HorizontalSymmetry As Boolean)
	Private CVS(6) As B4XCanvas
	Private CVSPanel(6) As B4XView
	Private CVSProxy(6) As BitmapCreator
	Public const MAX_SIZE_FOR_ANTIALIAS As Int = 300
	Private WorkingSpace As BitmapCreator
	Private X2 As X2Utils
	Private MAX_SIZE_OF_ALL_COMPRESSEDBCS As Int = 30 * 1024 * 1024
	Private TotalSize As Int
	Private Transform As B2Transform
	Private RectShape As B2PolygonShape
	Private OutputAABB As B2AABB
	Private TempId As Int
	Public Const TempPrefix As String = "~temp"
	Public CBCCache As InternalCompressedBCCache
End Sub

Public Sub Initialize (vX2 As X2Utils)
	CBCCache.Initialize
	CBCCache.ColorsMap.Initialize
	cache.Initialize
	WorkingSpace.Initialize(300, 300)
	Dim b(WorkingSpace.SAME_COLOR_LENGTH_FOR_CACHE * 4 * WorkingSpace.MAX_SAME_COLOR_SIZE + 4) As Byte
	CBCCache.mBuffer = b
	X2 = vX2
	Transform.Initialize
	RectShape.Initialize
	OutputAABB.Initialize
End Sub

Public Sub GetTempName As String
	TempId = TempId + 1
	Return TempPrefix & TempId
End Sub

'Returns a vector with the graphic size in meters.
Public Sub GetGraphicSizeMeters (Name As String, Index As Int) As B2Vec2
	Dim data As X2SpriteGraphicData = cache.Get(Name.ToLowerCase)
	Dim bc As BitmapCreator = data.OriginalBCs.Get(Index)
	Dim vec As B2Vec2
	vec.X = bc.mWidth / X2.mBCPixelsPerMeter
	vec.Y = bc.mHeight / X2.mBCPixelsPerMeter
	Return vec
End Sub
'Returns the number of graphic frames mapped to the given name.
Public Sub GetGraphicsCount(Name As String) As Int
	Dim data As X2SpriteGraphicData = cache.Get(Name.ToLowerCase)
	If data = Null Then Return 0
	Return data.OriginalBCs.Size
End Sub

'Name - Key
'X2ScaledBitmaps - List or array with the ScaledBitmaps objects.
Public Sub PutGraphic(Name As String, X2ScaledBitmaps As List) As X2SpriteGraphicData
	Dim sb As X2ScaledBitmap = X2ScaledBitmaps.Get(0)
	Dim antialias As Boolean = sb.Bmp.Width / sb.Scale * sb.Bmp.Height / sb.Scale < 3000
	If Name.StartsWith(TempPrefix) Then antialias = False
	Return PutGraphic2(Name, X2ScaledBitmaps, antialias, 5)
End Sub
 
'Name - Key
'X2ScaledBitmaps - List or array with the ScaledBitmaps objects.
'AntiAlias - Whether to use antialiasing when rotating the images. Can have a large impact on performace.
'AngleInterval - Default value is 5 degrees. 
Public Sub PutGraphic2(Name As String, X2ScaledBitmaps As List, AntiAlias As Boolean, AngleInterval As Int) As X2SpriteGraphicData
	Log($"New graphic: ${Name}"$)
	Dim data As X2SpriteGraphicData
	data.Initialize
	data.Name = Name
	data.MapsOfCompressedBCs.Initialize
	data.OriginalBCs.Initialize
	data.LastUsed = DateTime.Now
	For Each sb As X2ScaledBitmap In X2ScaledBitmaps
		Dim bc As BitmapCreator
		bc.Initialize(sb.Bmp.Width / sb.Scale, sb.bmp.Height / sb.Scale)
		bc.CopyPixelsFromBitmap(sb.bmp)
		data.OriginalBCs.Add(bc)
		Dim m As Map
		m.Initialize
		Dim cbc As CompressedBC = bc.ExtractCompressedBC(bc.TargetRect, CBCCache)
		m.Put(0, cbc)
		data.MapsOfCompressedBCs.Add(m)
		data.SizeOfAllCompressed = data.SizeOfAllCompressed + cbc.mBuffer.Length
	Next
	data.AngleInterval = AngleInterval
	data.AntiAlias = AntiAlias
	If AntiAlias Then
		data.OriginalBMPs.Initialize
		data.OriginalBMPs.AddAll(X2ScaledBitmaps)
	End If
	cache.Put(data.Name.ToLowerCase, data)
	TotalSize = TotalSize + data.SizeOfAllCompressed
	Return data
End Sub

'Gets a graphic frame from the cache.
Public Sub GetGraphic(Name As String, Index As Int) As CompressedBC
	Return GetGraphic2(Name, Index, 0, False, False)
End Sub

Public Sub GetGraphic2(Name As String, Index As Int, Degrees As Int, FlipHorizontally As Boolean, FlipVertically As Boolean) As CompressedBC
	Dim data As X2SpriteGraphicData = cache.Get(Name.ToLowerCase)
	If data = Null Then
		Log($"Error: graphic not found: ${Name}"$)
	End If
	data.LastUsed = DateTime.Now
	Dim m As Map = data.MapsOfCompressedBCs.Get(Index)
	Degrees = Degrees Mod 360
	If Degrees < 0 Then Degrees = 360 + Degrees
	Dim delta As Int = Degrees Mod data.AngleInterval
	If delta > data.AngleInterval / 2 Then
		Degrees = (Degrees - delta + data.AngleInterval) Mod 360
	Else
		Degrees = Degrees - delta
	End If
	Dim key As Int = Degrees
	If FlipHorizontally Then key = key + 1000
	If FlipVertically Then key = key + 2000
	If m.ContainsKey(key) Then Return m.Get(key)
	'Dim n As Long = DateTime.Now
	Dim TempBC As CompressedBC
	If FlipHorizontally <> FlipVertically Then
		Dim NonFlippedCbc As CompressedBC = GetGraphic2(Name, Index, -Degrees, False, False)
		TempBC = WorkingSpace.FlipCompressedBitmap(NonFlippedCbc, FlipHorizontally, FlipVertically)
	Else If FlipHorizontally And FlipVertically Then
		TempBC = GetGraphic2(Name, Index, Degrees + 180, False, False)
	Else If Degrees > 90 And (data.VerticalSymmetry Or data.HorizontalSymmetry) Then
		Dim quarter As Int = Degrees / 90
		Select quarter
			Case 1 '90 - 180
				TempBC = GetGraphic2(Name, Index, 180 - Degrees, False, False)
				TempBC = WorkingSpace.FlipCompressedBitmap(TempBC,Not(data.HorizontalSymmetry) , data.HorizontalSymmetry)
			Case 2 '180 - 270
				TempBC = GetGraphic2(Name, Index, Degrees - 180, False, False)
				TempBC = WorkingSpace.FlipCompressedBitmap(TempBC, True, True)
			Case 3 '270 - 360
				TempBC = GetGraphic2(Name, Index, 360 - Degrees, False, False)
				TempBC = WorkingSpace.FlipCompressedBitmap(TempBC, Not(data.VerticalSymmetry) , data.VerticalSymmetry)
		End Select
	Else If Degrees >= 180 Then
		TempBC = GetGraphic2(Name, Index, Degrees - 180, False, False)
		TempBC = WorkingSpace.FlipCompressedBitmap(TempBC, True, True)
	Else	
		Dim OriginalBC As BitmapCreator = data.OriginalBCs.Get(Index)
		Dim r As B4XRect = FindRotatedRect(OriginalBC, Degrees)
		If data.AntiAlias = False Or r.Width >= MAX_SIZE_FOR_ANTIALIAS Or r.Height >= MAX_SIZE_FOR_ANTIALIAS Then
			Dim dt As DrawTask = WorkingSpace.CreateDrawTask(OriginalBC, OriginalBC.TargetRect, r.Width / 2, r.Height / 2, True)
			dt.Degrees = Degrees
			If r.Width > WorkingSpace.mWidth Or r.Height > WorkingSpace.mHeight Then
				Log("Increasing WorkingSpace size.")
				Dim WorkingSpace As BitmapCreator
				WorkingSpace.Initialize(r.Width * 1.4, r.Height * 1.4)
			Else
				WorkingSpace.FillRect(xui.Color_Transparent, r)
			End If
			WorkingSpace.DrawBitmapCreatorTransformed(dt)
			TempBC = WorkingSpace.ExtractCompressedBC(r, CBCCache)
		Else
			TempBC = DrawRotatedBitmapWithCanvas(data.OriginalBMPs.Get(Index), Degrees, r)
		End If
	End If
	
	m.Put(key, TempBC)
	'Log($"Create new graphic: ${Name} - ${Degrees} ${FlipHorizontally} / ${FlipVertically} (${ (DateTime.Now - n)})"$)
	data.SizeOfAllCompressed = data.SizeOfAllCompressed + TempBC.mBuffer.Length
	TotalSize = TotalSize + TempBC.mBuffer.Length
	If TotalSize > MAX_SIZE_OF_ALL_COMPRESSEDBCS Then
		TrimCache	
	End If
	'Log(TotalSize & ". " & CBCCache.ColorsMap.Size)
	Return TempBC
End Sub

Private Sub TrimCache
	Log("Trim Cache")
	Log($"Before: ${TotalSize}"$)
	Dim dates As List
	dates.Initialize
	For Each data As X2SpriteGraphicData In cache.Values
		If data.SizeOfAllCompressed > 0 Then dates.Add(data)
	Next
	dates.SortType("LastUsed", True)
	For i = 0 To dates.Size / 2
		Dim data As X2SpriteGraphicData = dates.Get(i)
		For Each m As Map In data.MapsOfCompressedBCs
			m.Clear
		Next
		TotalSize = TotalSize - data.SizeOfAllCompressed
		data.SizeOfAllCompressed = 0
	Next
	Log($"After: ${TotalSize}"$)
End Sub

'Remove graphics from the cache. Call when the graphic is no longer needed.
Public Sub RemoveGraphics (Name As String)
	Log("Remove graphic: " & Name)
	Dim data As X2SpriteGraphicData = cache.Get(Name.ToLowerCase)
	TotalSize = TotalSize - data.SizeOfAllCompressed
	cache.Remove(Name.ToLowerCase)
End Sub

'Creates a DrawTask.
Public Sub GetDrawTask (Name As String, Index As Int, Degrees As Int, FlipH As Boolean, FlipV As Boolean, TargetX As Int, TargetY As Int) As DrawTask
	Dim sprite As CompressedBC = GetGraphic2(Name, Index, Degrees, FlipH, FlipV)
	Dim dt As DrawTask = WorkingSpace.CreateDrawTask(sprite, Null, 0, 0, False)
	dt.TargetX = TargetX - sprite.mWidth / 2
	dt.TargetY = TargetY - sprite.mHeight / 2
	dt.SrcRect = sprite.TargetRect
	dt.IsCompressedSource = True
	Return dt
End Sub

'Returns a square canvas larger than the specified size. Canvas.TargetView.Tag will return a BitmapCreator with the same non-smoothed size.
'The canvas size is multiplied by X2.BmpSmoothScale. 
'Maximum size is MAX_SIZE_FOR_ANTIALIAS (200).
Public Sub GetCanvas(Size As Int) As B4XCanvas
	Dim interval As Int = MAX_SIZE_FOR_ANTIALIAS / CVSPanel.Length
	Dim i As Int = Size / interval
	If CVSPanel(i).IsInitialized = False Then
		Dim MaxSize = (i + 1) * interval As Int
		If xui.IsB4J Then
			CVSPanel(i) = xui.CreatePanel("")
		Else
			Dim iv As ImageView
			iv.Initialize("")
			CVSPanel(i) = iv
		End If
		CVSPanel(i).SetLayoutAnimated(0, 0, 0, MaxSize * X2.BmpSmoothScale, _
			 MaxSize * X2.BmpSmoothScale)
		CVS(i).Initialize(CVSPanel(i))
		CVSProxy(i).Initialize(MaxSize, MaxSize)
		CVSPanel(i).Tag = CVSProxy(i)
	End If
	Return CVS(i)
End Sub

Private Sub DrawRotatedBitmapWithCanvas (SB As X2ScaledBitmap, Degrees As Int, r As B4XRect) As CompressedBC
	Dim c As B4XCanvas = GetCanvas(Max(r.Width, r.Height))
	Dim dest As B4XRect
	dest.Initialize(0, 0, 0, 0)
	Dim BmpWidth As Float = SB.bmp.Width / SB.Scale
	Dim BmpHeight As Float = SB.bmp.Height / SB.Scale
	dest.Left = (r.Width / 2 -  BmpWidth / 2) * X2.BmpSmoothScale
	dest.Top = (r.Height / 2 - BmpHeight / 2) * X2.BmpSmoothScale
	dest.Width = BmpWidth  * X2.BmpSmoothScale
	dest.Height= BmpHeight * X2.BmpSmoothScale
	c.ClearRect(c.TargetRect)
	c.DrawBitmapRotated(SB.bmp, dest, Degrees)
'	Dim n As Long = DateTime.Now
	Dim b As B4XBitmap = c.CreateBitmap
'	Dim CreateBitmapTime As Int = DateTime.Now - n
'	n = DateTime.Now
	Dim bc As BitmapCreator = c.TargetView.Tag
	bc.CopyPixelsFromBitmap(b)
'	Dim CopyPixelsTime As Int = DateTime.Now - n
	'Log($"CreateBitmap: ${CreateBitmapTime}, CopyPixels: ${CopyPixelsTime}"$)
	Return bc.ExtractCompressedBC(r, CBCCache)
End Sub

Private Sub FindRotatedRect(Input As BitmapCreator, Degrees As Int) As B4XRect
	Transform.Angle = X2.DegreesToB2Angle(Degrees)
	RectShape.SetAsBox(Input.mWidth / 2, Input.mHeight / 2)
	RectShape.ComputeAABB(OutputAABB, Transform)
	Dim r As B4XRect
	r.Initialize(0, 0, Ceil(OutputAABB.TopRight.X - OutputAABB.BottomLeft.X), Ceil(OutputAABB.TopRight.Y - OutputAABB.BottomLeft.Y))
	Return r	
End Sub
